@@ -1,70 +1,112 @@
-import os as _os
-import requests
-import mimetypes
-from urllib.parse import urlparse
-from urllib.request import urlretrieve
 import math
+import mimetypes
+import os as _os
+import traceback
+from queue import Queue, Empty
+from threading import Thread, Event
+from typing import Optional, List
+from urllib.parse import urlparse
+
+import requests
+from requests.adapters import HTTPAdapter, Retry
 
-from flytekit.exceptions.user import FlyteUserException
 from flytekit.configuration import DataConfig, LatchConfig
 from flytekit.core.data_persistence import DataPersistence, DataPersistencePlugins
-from queue import Queue
-from threading import Thread
-from typing import Optional
+from flytekit.exceptions.user import FlyteUserException
 
 """
 Thread Utils For Downloading Directories
 """
+
+
 class ThreadPool:
-    """ Pool of threads consuming tasks from a queue """
+    """Pool of threads consuming tasks from a queue"""
 
     def __init__(self, num_threads):
-        self.tasks = Queue(num_threads)
+        self.tasks: Queue = Queue(num_threads)
+        self.threads: List[Worker] = []
         for _ in range(num_threads):
-            Worker(self.tasks)
+            self.threads.append(Worker(self.tasks))
+
+    def start(self):
+        for t in self.threads:
+            t.start()
+
+    def join(self):
+        for t in self.threads:
+            t.join()
 
     def add_task(self, func, *args, **kargs):
-        """ Add a task to the queue """
+        """Add a task to the queue"""
         self.tasks.put((func, args, kargs))
 
     def map(self, func, args_list):
-        """ Add a list of tasks to the queue """
+        """Add a list of tasks to the queue"""
         for args in args_list:
             self.add_task(func, args)
 
     def wait_completion(self):
-        """ Wait for completion of all the tasks in the queue """
-        self.tasks.join()
+        """Wait for completion of all the tasks in the queue"""
+        try:
+            self.tasks.join()
+        finally:
+            for t in self.threads:
+                t.join()
 
 class Worker(Thread):
-    """ Thread executing tasks from a given tasks queue """
+    """Thread executing tasks from a given tasks queue"""
 
-    def __init__(self, tasks):
+    def __init__(self, tasks: Queue):
         Thread.__init__(self)
         self.tasks = tasks
         self.daemon = True
-        self.start()
 
     def run(self):
         while True:
-            func, args, kargs = self.tasks.get()
             try:
-                func(*args, **kargs)
-            except Exception as e:
-                # An exception happened in this thread
-                print(e)
+                func, args, kargs = self.tasks.get_nowait()
+            except Empty:
+                break
+
+            try:
+                attempts = 0
+                while True:
+                    try:
+                        func(*args, **kargs)
+                        break
+                    except ConnectionResetError as e:
+                        attempts += 1
+                        if attempts >= 5:
+                            raise e
+            except Exception:
+                traceback.print_exc()
             finally:
-                # Mark this task as done, whether an exception happened or not
                 self.tasks.task_done()
 
+def urlretrieve(url, path):
+    with open(path, 'wb') as f:
+        r = _session.get(url, stream=True)
+
+        for chunk in r.iter_content(1024):
+            f.write(chunk)
+
 def get(args):
     urlretrieve(args[0], args[1])
 
+
 def _enforce_trailing_slash(path: str):
     if path[-1] != "/":
         path += "/"
     return path
 
+
+_session = requests.Session()
+
+_adapter = HTTPAdapter(max_retries=Retry(status_forcelist=[429, 500, 502, 503, 504], backoff_factor=1))
+_session.mount("https://", _adapter)
+_session.mount("http://", _adapter)
+
+
 class LatchPersistence(DataPersistence):
     PROTOCOL = "latch://"
 
@@ -99,10 +141,13 @@ class LatchPersistence(DataPersistence):
         if not remote_path.startswith(self.PROTOCOL):
             raise ValueError(f"expected a Latch URL (latch://...): {remote_path}")
 
-        r = requests.post(self._latch_endpoint + "/api/object-exists-at-url", json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")})
+        r = _session.post(
+            self._latch_endpoint + "/api/object-exists-at-url",
+            json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")},
+        )
         if r.status_code != 200:
             raise FlyteUserException("failed to check if object exists at url `{}`".format(remote_path))
-        
+
         return r.json()["exists"]
 
     def download_directory(self, remote_path: str, local_path: str) -> bool:
@@ -110,7 +155,10 @@ class LatchPersistence(DataPersistence):
         :param str remote_path: remote latch:// path
         :param str local_path: directory to copy to
         """
-        r = requests.post(self._latch_endpoint + "/api/get-presigned-urls-for-dir", json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")})
+        r = _session.post(
+            self._latch_endpoint + "/api/get-presigned-urls-for-dir",
+            json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")},
+        )
         if r.status_code != 200:
             raise FlyteUserException("failed to download `{}`".format(remote_path))
 
@@ -127,6 +175,7 @@ class LatchPersistence(DataPersistence):
 
         pool = ThreadPool(100)
         pool.map(get, task_tuples)
+        pool.start()
         pool.wait_completion()
         return True
 
@@ -135,12 +184,15 @@ class LatchPersistence(DataPersistence):
         :param str remote_path: remote latch:// path
         :param str local_path: directory to copy to
         """
-        r = requests.post(self._latch_endpoint + "/api/get-presigned-url", json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")})
+        r = _session.post(
+            self._latch_endpoint + "/api/get-presigned-url",
+            json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")},
+        )
         if r.status_code != 200:
             raise FlyteUserException("failed to get presigned url for `{}`".format(remote_path))
 
         url = r.json()["url"]
-        urlretrieve(url, local_path)
+        get((url, local_path))
         return _os.path.exists(local_path)
 
     @staticmethod
@@ -155,27 +207,43 @@ class LatchPersistence(DataPersistence):
         if content_type is None:
             content_type = "application/octet-stream"
 
-        r = requests.post(endpoint + "/api/begin-upload", json={"object_url": to_path, "nrof_parts": nrof_parts, "content_type": content_type, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")})
+        r = _session.post(
+            endpoint + "/api/begin-upload",
+            json={
+                "object_url": to_path,
+                "nrof_parts": nrof_parts,
+                "content_type": content_type,
+                "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID"),
+            },
+        )
         if r.status_code != 200:
             raise FlyteUserException("failed to get presigned upload urls for `{}`".format(to_path))
-        
+
         data = r.json()
         presigned_urls = data["urls"]
         upload_id = data["upload_id"]
         f = open(file_path, "rb")
-        parts=[]
+        parts = []
         for key, val in presigned_urls.items():
             blob = f.read(chunk_size)
-            r = requests.put(val, data=blob)
+            r = _session.put(val, data=blob)
             if r.status_code != 200:
                 print(r.status_code)
                 print(r.text)
                 print(r.headers)
                 raise RuntimeError("failed to upload part `{}` of file `{}`".format(key, file_path))
-            etag = r.headers['ETag']
-            parts.append({'ETag': etag, 'PartNumber': int(key) + 1})
-        
-        r = requests.post(endpoint + "/api/complete-upload", json={"upload_id": upload_id, "parts": parts, "object_url": to_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")})
+            etag = r.headers["ETag"]
+            parts.append({"ETag": etag, "PartNumber": int(key) + 1})
+
+        r = _session.post(
+            endpoint + "/api/complete-upload",
+            json={
+                "upload_id": upload_id,
+                "parts": parts,
+                "object_url": to_path,
+                "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID"),
+            },
+        )
         if r.status_code != 200:
             raise RuntimeError("failed to complete upload for `{}`".format(to_path))
         return True
@@ -210,9 +278,10 @@ class LatchPersistence(DataPersistence):
 
         pool = ThreadPool(100)
         pool.map(LatchPersistence.__upload, task_tuples)
+        pool.start()
         pool.wait_completion()
         return True
-    
+
     def construct_path(self, add_protocol: bool, add_prefix: bool, *paths: str) -> str:
         paths = list(paths)  # make type check happy
         if add_prefix:
@@ -221,14 +290,14 @@ class LatchPersistence(DataPersistence):
         if add_protocol:
             return f"{self.PROTOCOL}{path}"
         return path
-    
+
     def get(self, from_path: str, to_path: str, recursive: bool = False):
         if not from_path.startswith(self.PROTOCOL):
             raise ValueError(f"expected a Latch URL (latch://...): {from_path}")
 
         if recursive:
             return self.download_directory(from_path, to_path)
-        
+
         return self.download(from_path, to_path)
 
     def put(self, from_path: str, to_path: str, recursive: bool = False):
@@ -237,7 +306,8 @@ class LatchPersistence(DataPersistence):
 
         if recursive:
             return self.upload_directory(from_path, to_path)
-        
+
         return self.upload(from_path, to_path)
 
+
 DataPersistencePlugins.register_plugin(LatchPersistence.PROTOCOL, LatchPersistence)
